//  $Id: memory.cc,v 1.2 2004/03/25 04:19:07 vickery Exp $
/*
 *    Implements the main memory for the ARC simulator.
 *
 *    C. Vickery
 *    Spring 2004
 *
 *  Copyright (c) 2004, Queens College of the City University
 *  of New York.  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or
 *  without modification, are permitted provided that the
 *  following conditions are met:
 *
 *      * Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the
 *        following disclaimer.
 * 
 *      * Redistributions in binary form must reproduce the
 *        above copyright notice, this list of conditions and
 *        the following disclaimer in the documentation and/or
 *        other materials provided with the distribution.  
 * 
 *      * Neither the name of Queens College of CUNY
 *        nor the names of its contributors may be used to
 *        endorse or promote products derived from this
 *        software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 *  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *    $Log: memory.cc,v $
 *    Revision 1.2  2004/03/25 04:19:07  vickery
 *    Initial version.  Implements memory_init(), memory_read(), and
 *    memory_write().
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "simulator.h"

//  memory
//  ------------------------------------------------------------------
static uint32_t *memory = 0;
static uint16_t last_address = 0;

//  memory_init()
//  ------------------------------------------------------------------
/*
 *    Initialize a memory of size N words for simulation.  Release
 *    previous allocation if there was one.
 */
  void
  memory_init( uint16_t num_words )
  {
    if ( memory != 0 )
    {
      free( memory );
    }
    memory = (uint32_t *) malloc( num_words * 4 );
    if ( 0 == memory )
    {
      perror( "Allocating simulated memory" );
      exit( 1 );
    }
    last_address = (num_words * 4 ) - 1;
  }

//  memory_read()
//  ------------------------------------------------------------------
/*
 *    Read a word of memory.
 */
  uint32_t memory_read( uint32_t address )
  {
    //  Validate the request
    if ( 0 == memory )
    {
      fprintf( stderr, "memory_read: Attempt to read from memory "
                                 "before initializing simulator.\n" );
      exit( 1 );
    }

    //   Check for "null pointer"
    if ( 0 == address )
    {
      fprintf( stderr, "memory_write: Write to address 0x00000000 "
                                " not allowed.  Request ignored.\n" );
      return 0;
    }

     if ( (address + 3) > last_address )
    {
      fprintf( stderr, "memory_read: Read address %08X is above end "
             "of simulated memory (%08X).\n", address, last_address );
      exit( 1 );
    }

    if ( (address & 3) != 0 )
    {
      fprintf( stderr, "memory_read: Warning: Read address %08X is "
                                            "not a word address.\n" );
    }

    //  Perform the actual read operation
    return memory[ address >> 2 ];
  }


//  memory_write()
//  ------------------------------------------------------------------
/*
 *    Write a word of memory.
 */
  void memory_write( uint32_t address, uint32_t data )
  {
    //  Validate the request
    if ( 0 == memory )
    {
      fprintf( stderr, "memory_write: Attempt to write to memory "
                                 "before initializing simulator.\n" );
      exit( 1 );
    }

    //   Check for "null pointer"
    if ( 0 == address )
    {
      fprintf( stderr, "memory_write: Write to address 0x00000000 "
                                " not allowed.  Request ignored.\n" );
      return;
    }

    if ( (address + 3) > last_address )
    {
      fprintf( stderr, "memory_write: Write address %08X is above "
         "end of simulated memory (%08X).\n", address, last_address );
      exit( 1 );
    }

    if ( (address & 3) != 0 )
    {
      fprintf( stderr, "memory_write: Warning: Write address %08X is "
                                  "not a word address.\n", address  );
    }

    //  Perform the actual write operation
    memory[ address >> 2 ] = data;
    return;
  }

